import io
import operator
from pathlib import Path, PosixPath
from collections import Counter
from datetime import datetime
import re
import csv

import numpy as np

from tqdm import tqdm

from logfile_parser import Parser
from omero.gateway import BlitzGateway, TagAnnotationWrapper


class OmeroExplorer:
    def __init__(self, host, user, password, min_date=(2020, 6, 1)):
        self.conn = BlitzGateway(user, password, host=host)
        self.conn.connect()

        self.min_date = min_date
        self.backups = {}
        self.removed = []

    @property
    def cache(self):
        if not hasattr(self, "_cache"):
            self._cache = {v.id: get_annotsets(v) for v in self.dsets}
        return self._cache

    @property
    def raw_log(self):
        return {k: v["log"] for k, v in self.cache.items()}

    @property
    def raw_log_end(self):
        if not hasattr(self, "_raw_log_end"):
            self._raw_log_end = {d.id: get_logfile(d) for d in self.dsets}
        return self._raw_log_end

    @property
    def log(self):
        return {k: parse_annot(v, "log") for k, v in self.raw_log.items()}

    @property
    def raw_acq(self):
        return {k: v["acq"] for k, v in self.cache.items()}

    @property
    def acq(self):
        return {k: parse_annot(v, "acq") for k, v in self.raw_acq.items()}

    def load(self, min_id=18000, min_date=None):
        """
        :min_id: int
        :min_date: tuple
        """
        if min_date is None:
            min_date = self.min_date
        self._dsets_bak = [
            d for d in self.conn.getObjects("Dataset") if d.getId() > min_id
        ]

        if min_date:
            if len(min_date) < 3:
                min_date = min_date + tuple([1 for i in range(3 - len(min_date))])
            min_date = datetime(*min_date)

            # sort by dates
            dates = [d.getDate() for d in self._dsets_bak]
            self._dsets_bak[:] = [a for b, a in sorted(zip(dates, self._dsets_bak))]

            self._dsets_bak = [d for d in self._dsets_bak if d.getDate() >= min_date]

        self.dsets = self._dsets_bak
        self.n_dsets

    def dset(self, n):
        try:
            return [x for x in self.dsets if x.id == n][0]
        except:
            return

    def channels(self, setkey, present=True):
        """
        :setkey: str indicating a set of channels
        :present: bool indicating whether the search should or shold not be in the dset
        """
        self.dsets = [
            v for v in self.acqs.values() if present == has_channels(v, setkey)
        ]
        self.n_dsets

    def update_cache(self):
        if not hasattr(self, "acq") or not hasattr(self, "log"):
            for attr in ["acq", "log"]:
                print("Updating raw ", attr)
                setattr(
                    self,
                    "raw_" + attr,
                    {v.id: get_annotsets(v)[attr] for v in self.dsets},
                )
                setattr(
                    self,
                    attr,
                    {
                        v.id: parse_annot(getattr(self, "raw_" + attr)[v.id], attr)
                        for v in self.dsets
                    },
                )
        else:

            for attr in ["acq", "log", "raw_acq", "raw_log"]:
                setattr(
                    self, attr, {i.id: getattr(self, attr)[i.id] for i in self.dsets}
                )

    @property
    def dsets(self):
        if not hasattr(self, "_dsets"):
            self._dsets = self.load()

        return self._dsets

    @dsets.setter
    def dsets(self, dsets):
        if hasattr(self, "_dsets"):
            if self._dsets is None:
                self._dsets = []
            self.removed += [
                x for x in self._dsets if x.id not in [y.id for y in dsets]
            ]

        self._dsets = dsets

    def tags(self, tags, present=True):
        """
        :setkey str tags to filter
        """
        if type(tags) is not list:
            tags = [str(tags)]

        self.dsets = [v for v in self.dsets if present == self.has_tags(v, tags)]
        self.n_dsets

    @property
    def all_tags(self):
        if not hasattr(self, "_tags"):
            self._tags = {
                d.id: [
                    x.getValue()
                    for x in d.listAnnotations()
                    if isinstance(x, TagAnnotationWrapper)
                ]
                for d in self.dsets
            }
        return self._tags

    def get_timepoints(self):
        self.image_wrappers = {d.id: list(d.listChildren())[0] for d in self.dsets}

        return {k: i.getSizeT() for k, i in self.image_wrappers.items()}

    def timepoints(self, n, op="greater"):
        "Filter experiments using the number of timepoints"
        op = operator.gt if op == "greater" else operator.le
        self._timepoints = self.get_timepoints()

        self.dsets = [v for v in tqdm(self.dsets) if op(self._timepoints[v.id], n)]

    def microscope(self, microscope):
        self.microscopes = {
            dset.id: self.get_microscope(self.log[dset.id]) for dset in self.dsets
        }

        self.n_dsets

    def get_microscope(self, parsed_log):
        return parsed_log["microscope"]

    def reset(self, backup_id=None):
        self.dsets = self.backups.get(backup_id, self._dsets_bak)
        self.n_dsets

    def backup(self, name):
        self.backups[name] = self.dsets

    def reset_backup(self, name):
        self.dsets = self.backups[name]

    def cExperiment(self, present=True):
        self.dsets = [
            v
            for v in self.dsets
            if present
            * sum(
                [
                    "cExperiment" in x.getFileName()
                    for x in v.listAnnotations()
                    if hasattr(x, "getFileName")
                ]
            )
        ]
        self.n_dsets

    @staticmethod
    def is_complete(logfile):
        return logfile.endswith("Experiment completed\r\r\n")

    @staticmethod
    def contains_regex(logfile):
        pass
        # return re.

    def tiler_cells(self, present=True):
        self.__dsets = [v for v in self.dsets if present == tiler_cells_load(v)]

    @property
    def n_dsets(self):
        print("{} datasets.".format(len(self.dsets)))

    @property
    def desc(self):
        for d in self.dsets:
            print(
                "{}\t{}\t{}\t{}".format(
                    d.getDate().strftime("%x"),
                    d.getId(),
                    d.getName(),
                    d.getDescription(),
                )
            )

    @property
    def ids(self):
        return [d.getId() for d in self.dsets]

    # @property
    # def acqs(self):
    #     if not hasattr(self, "_acqs") or len(self.__dict__.get("_acqs", [])) != len(
    #         self.dsets
    #     ):
    #         self._acqs = [get_annot(get_annotsets(d), "acq") for d in self.dsets]
    #     return self._acqs

    def get_ph_params(self):
        t = [
            {
                ch: [exp, v]
                for ch, exp, v in zip(j["channel"], j["exposure"], j["voltage"])
                if ch in {"GFPFast", "pHluorin405"}
            }
            for j in [i["channels"] for i in self.acqs]
        ]

        ph_param_pairs = [(tuple(x.values())) for x in t if np.all(list(x.values()))]

        return Counter([str(x) for x in ph_param_pairs])

    def find_duplicate_candidates(self, days_tol=2):
        # Find experiments with the same name or Aim and from similar upload dates
        # and group them for cleaning
        pass

    def group_by_date(tol=1):
        dates = [x.getDate() for x in self.dsets]
        distances = np.array(
            [[abs(convert_to_hours(a - b)) for a in dates] for b in dates]
        )
        return explore_booldiag(distances > tol, 0, [])

    @property
    def complete(self):
        self.completed = {k: self.is_complete(v) for k, v in self.raw_log_end.items()}
        self.dsets = [dset for dset in self.dsets if self.completed[dset.id]]
        return self.n_dsets

    def save(self, fname):
        with open(fname + ".tsv", "w") as f:
            writer = csv.writer(f, delimiter="\t")
            for d in self.dsets:
                writer.writerow(
                    [
                        d.getDate().strftime("%x"),
                        d.getId(),
                        d.getName(),
                        d.getDescription(),
                    ]
                )

    @property
    def positions(self):
        return {x.id: len(list(x.listChildren())) for x in self.dsets}

    def has_tags(self, d, tags):
        if set(tags).intersection(self.all_tags[d.id]):
            return True


def explore_booldiag(bool_field, current_position, cluster_start_end):
    # Recursively find the square clusters over the diagonal. Allows for duplicates
    # returns a list of tuples with the start, end of clusters
    if current_position < len(bool_field) - 1:
        elements = np.where(bool_field[current_position])
        if len(elements[0]) > 1:
            start = elements[0][0]
            end = elements[0][-1]
        else:
            start = elements[0][0]
            end = elements[0][0]

        cluster_start_end.append((start, end))
        return explore_square(bool_field, end + 1, cluster_start_end)
    else:
        return cluster_start_end
    _


def convert_to_hours(delta):
    total_seconds = delta.total_seconds()
    hours = int(total_seconds // 3600)
    return hours


class Argo(OmeroExplorer):
    def __init__(self,*args, **kwargs):
        super().__init__(*args,**kwargs)


def get_creds():
    return (
        "upload",
        "***REMOVED***",  # OMERO Password
    )


def list_files(dset):
    return {x for x in dset.listAnnotations() if hasattr(x, "getFileName")}


def annot_from_dset(dset, kind):
    v = [x for x in dset.listAnnotations() if hasattr(x, "getFileName")]
    infname = kind if kind is "log" else kind.title()
    try:
        acqfile = [x for x in v if x.getFileName().endswith(infname + ".txt")][0]
        decoded = list(acqfile.getFileInChunks())[0].decode("utf-8")
        acq = parse_annot(decoded, kind)
    except:
        return {}

    return acq


def check_channels(acq, channels, _all=True):
    I = set(acq["channels"]["channel"]).intersection(channels)

    condition = False
    if _all:
        if len(I) == len(channels):
            condition = True
    else:
        if len(I):
            condition = True

    return condition


def get_chs(exptype):
    exptypes = {
        "dual_ph": ("GFP", "pHluorin405", "mCherry"),
        "ph": ("GFP", "pHluorin405"),
        "dual": ("GFP", "mCherry"),
        "mCherry": ("mCherry"),
    }
    return exptypes[exptype]


def load_annot_from_cache(exp_id, cache_dir="cache/"):
    if type(cache_dir) is not PosixPath:
        cache_dir = Path(cache_dir)

    annot_sets = {}
    for fname in cache_dir.joinpath(exp_id).rglob("*.txt"):
        fmt = fname.name[:3]
        with open(fname, "r") as f:
            annot_sets[fmt] = f.read()

    return annot_sets


def get_annot(annot_sets, fmt):
    """
    Get parsed annotation file
    """
    str_io = annot_sets.get(fmt, None)
    return parse_annot(str_io, fmt)


def parse_annot(str_io, fmt):
    parser = Parser("multiDGUI_" + fmt + "_format")
    return parser.parse(io.StringIO(str_io))


def get_log_date(annot_sets):
    log = get_annot(annot_sets, "log")
    return log.get("date", None)


def get_log_microscope(annot_sets):
    log = get_annot(annot_sets, "log")
    return log.get("microscope", None)


def get_annotsets(dset):
    annot_files = [
        annot.getFile() for annot in dset.listAnnotations() if hasattr(annot, "getFile")
    ]
    annot_sets = {
        ftype[:-4].lower(): annot
        for ftype in ("log.txt", "Acq.txt", "Pos.txt")
        for annot in annot_files
        if ftype in annot.getName()
    }
    annot_sets = {
        key: list(a.getFileInChunks())[0].decode("utf-8")
        for key, a in annot_sets.items()
    }
    return annot_sets


# def has_tags(d, tags):
#     if set(tags).intersection(annot_from_dset(d, "log").get("omero_tags", [])):
#         return True


def load_acq(dset):
    try:
        acq = annot_from_dset(dset, kind="acq")
        return acq
    except:
        print("dset", dset.getId(), " failed acq loading")
        return False


def has_channels(dset, exptype):
    acq = load_acq(dset)
    if acq:
        return check_channels(acq, get_chs(exptype))
    else:
        return


def get_id_from_name(exp_name, conn=None):
    if conn is None:
        conn = BlitzGateway(*get_creds(), host="islay.bio.ed.ac.uk", port=4064)

    if not conn.isConnected():
        conn.connect()

    cand_dsets = [
        d
        for d in conn.getObjects("Dataset")  # , opts={'offset': 10600,
        #      'limit':500})
        if exp_name in d.name
    ]  # increase the offset for better speed

    # return cand_dsets
    if len(cand_dsets) > 1:
        # Get date and try to find it using date and microscope name and date

        # found = []
        # for cand in cand_dsets:
        #     annot_sets = get_annotsets(cand)
        #     date = get_log_date(annot_sets)
        #     microscope = get_log_microscope(annot_sets)
        #     if date==date_name and microscope == microscope_name:
        #         found.append(cand)

        # if True:#len(found)==1:
        #     return best_cand.id#best_cand = found[0]
        if True:

            print("Multiple options found. Selecting the one with most children")

            max_dset = np.argmax(
                [
                    len(list(conn.getObject("Dataset", c.id).listChildren()))
                    for c in cand_dsets
                ]
            )

            best_cand = cand_dsets[max_dset]

            return best_cand.id
    elif len(cand_dsets) == 1:
        return cand_dsets[0].id


# Custom functions
def compare_dsets_voltages_exp(dsets):
    a = {}
    for d in dsets:
        try:
            acq = annot_from_dset(d, kind="acq")["channels"]
            a[d.getId()] = {
                k: (v, e)
                for k, v, e in zip(acq["channel"], acq["voltage"], acq["exposure"])
            }

        except:
            print(d, "didnt work")

    return a


def get_logfile(dset):
    annot_file = [
        annot.getFile()
        for annot in dset.listAnnotations()
        if hasattr(annot, "getFile") and annot.getFileName().endswith("log.txt")
    ][0]
    return list(annot_file.getFileInChunks())[-1].decode("utf-8")


# 19920 -> 19300/19310
#



import numpy as np
from pathlib import Path
from extraction.core.extractor import Extractor
from extraction.core.parameters import Parameters
from extraction.core.functions.defaults import get_params

params = Parameters(**get_params("batman_ph_dual_fast"))
# ext = Extractor(params, source=19918)  # 19831
ext = Extractor(params, source=19831)
ext.load_tiler()
self = ext
# s=self.extract_exp(tree={'general':{None:['area']}, 'GFPFast':{np.maximum:['median','mean']}},poses=self.expt.positions[:2], tps=[0,1], stg='df')
s = self.extract_exp()
# # import cProfile
# # profile = cProfile.Profile()
# # profile.enable()
# # ext.change_position(ext.expt.positions[1])
# # tracks = self.extract_pos(
# #     tree={('general'):{None:  # Other metrics can be used
# #                        [tidy_metric]}})#['general',None,'area']

# # profile.disable()
# # import pstats
# # ps = pstats.Stats(profile)
# # ps.sort_stats('cumulative')
# # ps.print_stats()



import git
import pkg_resources


def get_sha():
    repo = git.Repo(search_parent_directories=True)
    sha = repo.head.object.hexsha
    return sha


def get_version(pkg="extraction"):
    return pkg_resources.require(pkg)[0].version



import numpy as np


def div0(a, b, fill=0):
    """a / b, divide by 0 -> `fill`
    div0( [-1, 0, 1], 0, fill=np.nan) -> [nan nan nan]
    div0( 1, 0, fill=np.inf ) -> inf
    """
    with np.errstate(divide="ignore", invalid="ignore"):
        c = np.true_divide(a, b)
    if np.isscalar(c):
        return c if np.isfinite(c) else fill
    else:
        c[~np.isfinite(c)] = fill
        return c



from core.experiment import Experiment
from core.segment import Tiler
expt = Experiment.from_source(19310, #Experiment ID on OMERO
                                'upload', #OMERO Username
                                '***REMOVED***', #OMERO Password
                                'islay.bio.ed.ac.uk', #OMERO host
                                port=4064 #This is default
                                )


# Load whole position
img=expt[0,0,:,:,2]
plt.imshow(img[0,0,...,0]); plt.show()

# Manually get template
tilesize=117
x0=827
y0=632
trap_template = img[0,0,x0:x0+tilesize,y0:y0+tilesize,0]
plt.imshow(trap_template); plt.show()

tiler = Tiler(expt, template = trap_template)

# Load images (takes about 5 mins)
trap_tps = tiler.get_traps_timepoint(0, tile_size=117, z=[2])

#Plot found traps
nrows, ncols = (5,5)
fig, axes = plt.subplots(nrows,ncols)
for i in range(nrows):
    for j in range(ncols):
        if i*nrows+j < trap_tps.shape[0]:
            axes[i,j].imshow(trap_tps[i*nrows+j,0,0,...,0])
plt.show()




"""
Pipeline and chaining elements.
"""
import logging
import os
from abc import ABC, abstractmethod
from typing import List
from pathlib import Path
import traceback

from itertools import groupby
import h5py
import yaml
from tqdm import tqdm
from time import perf_counter
from pathos.multiprocessing import Pool

import numpy as np
import pandas as pd
from scipy import ndimage

from aliby.experiment import MetaData
from aliby.haystack import initialise_tf
from aliby.baby_client import BabyRunner, BabyParameters
from aliby.tile.tiler import Tiler, TilerParameters
from aliby.io.omero import Dataset, Image
from agora.abc import ParametersABC, ProcessABC
from agora.io.writer import TilerWriter, BabyWriter, StateWriter
from agora.io.signal import Signal
from extraction.core.extractor import Extractor, ExtractorParameters
from extraction.core.functions.defaults import exparams_from_meta
from postprocessor.core.processor import PostProcessor, PostProcessorParameters
from postprocessor.compiler import ExperimentCompiler, PageOrganiser

logging.basicConfig(
    filename="aliby.log",
    filemode="w",
    format="%(name)s - %(levelname)s - %(message)s",
    level=logging.DEBUG,
)


class PipelineParameters(ParametersABC):
    def __init__(
        self, general, tiler, baby, extraction, postprocessing, reporting=None
    ):
        self.general = general
        self.tiler = tiler
        self.baby = baby
        self.extraction = extraction
        self.postprocessing = postprocessing
        self.reporting = reporting

    @classmethod
    def default(
        cls,
        general={},
        tiler={},
        baby={},
        extraction={},
        postprocessing={},
        reporting={},
    ):
        """
        Load unit test experiment
        :expt_id: Experiment id
        :directory: Output directory

        Provides default parameters for the entire pipeline. This downloads the logfiles and sets the default
        timepoints and extraction parameters from there.
        """
        expt_id = general.get("expt_id", 19993)
        directory = Path(general.get("directory", "../data"))
        with Dataset(int(expt_id), **general.get("server_info")) as conn:
            directory = directory / conn.unique_name
            if not directory.exists():
                directory.mkdir(parents=True)
                # Download logs to use for metadata
            conn.cache_logs(directory)
        meta = MetaData(directory, None).load_logs()
        tps = meta["time_settings/ntimepoints"][0]
        defaults = {
            "general": dict(
                id=expt_id,
                distributed=0,
                tps=tps,
                directory=str(directory),
                strain="",
                earlystop=dict(
                    min_tp=100,
                    thresh_pos_clogged=0.3,
                    thresh_trap_clogged=7,
                    ntps_to_eval=5,
                ),
            )
        }
        defaults["tiler"] = TilerParameters.default().to_dict()
        defaults["baby"] = BabyParameters.default().to_dict()
        defaults["extraction"] = exparams_from_meta(meta)
        defaults["postprocessing"] = PostProcessorParameters.default().to_dict()
        for k in defaults.keys():
            exec("defaults[k].update(" + k + ")")
        return cls(**{k: v for k, v in defaults.items()})

    def load_logs(self):
        parsed_flattened = parse_logfiles(self.log_dir)
        return parsed_flattened


class Pipeline(ProcessABC):
    """
    A chained set of Pipeline elements connected through pipes.
    Tiling, Segmentation,Extraction and Postprocessing should use their own default parameters.
    These can be overriden passing the key:value of parameters to override to a PipelineParameters class

    """

    def __init__(self, parameters: PipelineParameters, store=None):
        super().__init__(parameters)

        if store is None:
            store = self.parameters.general["directory"]
        self.store = store

    @classmethod
    def from_yaml(cls, fpath):
        # This is just a convenience function, think before implementing
        # for other processes
        return cls(parameters=PipelineParameters.from_yaml(fpath))

    @classmethod
    def from_existing_h5(cls, fpath):
        with h5py.File(fpath, "r") as f:
            pipeline_parameters = PipelineParameters.from_yaml(f.attrs["parameters"])

        return cls(pipeline_parameters)

    def run(self):
        # Config holds the general information, use in main
        # Steps holds the description of tasks with their parameters
        # Steps: all holds general tasks
        # steps: strain_name holds task for a given strain
        config = self.parameters.to_dict()
        expt_id = config["general"]["id"]
        distributed = config["general"]["distributed"]
        strain_filter = config["general"]["strain"]
        root_dir = config["general"]["directory"]
        root_dir = Path(root_dir)

        print("Searching OMERO")
        # Do all initialis
        with Dataset(int(expt_id), **self.general["server_info"]) as conn:
            image_ids = conn.get_images()
            directory = root_dir / conn.unique_name
            if not directory.exists():
                directory.mkdir(parents=True)
                # Download logs to use for metadata
            conn.cache_logs(directory)

        # Modify to the configuration
        self.parameters.general["directory"] = directory
        config["general"]["directory"] = directory

        # Filter TODO integrate filter onto class and add regex
        if isinstance(strain_filter, str):
            image_ids = {
                k: v for k, v in image_ids.items() if k.startswith(strain_filter)
            }
        elif isinstance(strain_filter, int):
            image_ids = {
                k: v for i, (k, v) in enumerate(image_ids.items()) if i == strain_filter
            }

        if distributed != 0:  # Gives the number of simultaneous processes
            with Pool(distributed) as p:
                results = p.map(lambda x: self.create_pipeline(x), image_ids.items())
            return results
        else:  # Sequential
            results = []
            for k, v in image_ids.items():
                r = self.create_pipeline((k, v))
                results.append(r)

    def create_pipeline(self, image_id):
        config = self.parameters.to_dict()
        name, image_id = image_id
        general_config = config["general"]
        session = None
        earlystop = general_config["earlystop"]
        try:
            directory = general_config["directory"]

            with Image(image_id, **self.general["server_info"]) as image:
                filename = f"{directory}/{image.name}.h5"
                meta = MetaData(directory, filename)
                from_start = True
                if (
                    not general_config.get("overwrite", False)
                    and Path(filename).exists()
                ):
                    try:
                        print(f"Existing file {filename} will be used.")
                        with h5py.File(filename, "r") as f:
                            tiler = Tiler.from_hdf5(image, filename)
                            s = Signal(filename)
                            process_from = (
                                f.attrs["last_processed"]
                                or s.get_raw("/general/None/extraction/volume").columns[
                                    -1
                                ]
                                or 0
                            )
                            # get state array
                            state_array = f.get("state_array", 0)
                        if process_from > 2:
                            processFalsefrom = process_from - 3
                            tiler.n_processed = process_from
                        from_start = False
                    except:
                        pass

                if from_start:  # New experiment or overwriting
                    try:
                        os.remove(filename)
                    except:
                        pass

                    process_from = 0
                    try:
                        meta.run()
                        meta.add_fields(
                            {"omero_id,": config["general"]["id"], "image_id": image_id}
                        )
                        tiler = Tiler.from_image(
                            image, TilerParameters.from_dict(config["tiler"])
                        )
                    except:
                        # Remove and try to run again?
                        pass

                writer = TilerWriter(filename)
                session = initialise_tf(2)
                runner = BabyRunner.from_tiler(
                    BabyParameters.from_dict(config["baby"]), tiler
                )
                bwriter = BabyWriter(filename)

                # Limit extraction parameters during run using the available channels in tiler
                av_channels = set((*tiler.channels, "general"))
                config["extraction"]["tree"] = {
                    k: v
                    for k, v in config["extraction"]["tree"].items()
                    if k in av_channels
                }
                config["extraction"]["sub_bg"] = av_channels.intersection(
                    config["extraction"]["sub_bg"]
                )

                av_channels_wsub = av_channels.union(
                    [c + "_bgsub" for c in config["extraction"]["sub_bg"]]
                )
                for op in config["extraction"]["multichannel_ops"]:
                    config["extraction"]["multichannel_ops"][op] = [
                        x
                        for x in config["extraction"]["multichannel_ops"]
                        if len(x[0]) == len(av_channels_wsub.intersection(x[0]))
                    ]
                config["extraction"]["multichannel_ops"] = {
                    k: v
                    for k, v in config["extraction"]["multichannel_ops"].items()
                    if len(v)
                }

                exparams = ExtractorParameters.from_dict(config["extraction"])
                ext = Extractor.from_tiler(exparams, store=filename, tiler=tiler)

                # RUN
                tps = general_config["tps"]
                frac_clogged_traps = 0
                for i in tqdm(
                    range(process_from, tps), desc=image.name, initial=process_from
                ):
                    if (
                        frac_clogged_traps < earlystop["thresh_pos_clogged"]
                        or i < earlystop["min_tp"]
                    ):
                        t = perf_counter()
                        trap_info = tiler.run_tp(i)
                        logging.debug(f"Timing:Trap:{perf_counter() - t}s")
                        t = perf_counter()
                        writer.write(trap_info, overwrite=[])
                        logging.debug(f"Timing:Writing-trap:{perf_counter() - t}s")
                        t = perf_counter()
                        seg = runner.run_tp(i)
                        logging.debug(f"Timing:Segmentation:{perf_counter() - t}s")
                        # logging.debug(
                        #     f"Segmentation failed:Segmentation:{perf_counter() - t}s"
                        # )
                        t = perf_counter()
                        bwriter.write(seg, overwrite=["mother_assign"])
                        logging.debug(f"Timing:Writing-baby:{perf_counter() - t}s")

                        t = perf_counter()
                        labels, masks = groupby_traps(
                            seg["trap"],
                            seg["cell_label"],
                            seg["edgemasks"],
                            tiler.n_traps,
                        )
                        tmp = ext.run(tps=[i], masks=masks, labels=labels)
                        logging.debug(f"Timing:Extraction:{perf_counter() - t}s")
                    else:  # Stop if more than X% traps are clogged
                        logging.debug(
                            f"EarlyStop:{earlystop['thresh_pos_clogged']*100}% traps clogged at time point {i}"
                        )
                        print(
                            f"Stopping analysis at time {i} with {frac_clogged_traps} clogged traps"
                        )
                        break

                    if (
                        i > earlystop["min_tp"]
                    ):  # Calculate the fraction of clogged traps
                        frac_clogged_traps = self.check_earlystop(filename, earlystop)
                        logging.debug(f"Quality:Clogged_traps:{frac_clogged_traps}")
                        print("Frac clogged traps: ", frac_clogged_traps)

                    meta.add_fields({"last_processed": i})
                # Run post processing
                post_proc_params = PostProcessorParameters.from_dict(
                    self.parameters.postprocessing
                ).to_dict()
                PostProcessor(filename, post_proc_params).run()

                return 1
        except Exception as e:  # bug in the trap getting
            logging.exception(
                f"Caught exception in worker thread (x = {name}):", exc_info=True
            )
            print(f"Caught exception in worker thread (x = {name}):")
            # This prints the type, value, and stack trace of the
            # current exception being handled.
            traceback.print_exc()
            print()
            raise e
        finally:
            if session:
                session.close()

        try:
            compiler = ExperimentCompiler(None, filepath)
            tmp = compiler.run()
            po = PageOrganiser(tmp, grid_spec=(3, 2))
            po.plot()
            po.save(fullpath / f"{directory}report.pdf")
        except Exception as e:
            print(e)

    def check_earlystop(self, filename, es_parameters):
        s = Signal(filename)
        df = s["/extraction/general/None/area"]
        frac_clogged_traps = (
            df[df.columns[-1 - es_parameters["ntps_to_eval"] : -1]]
            .dropna(how="all")
            .notna()
            .groupby("trap")
            .apply(sum)
            .apply(np.mean, axis=1)
            > es_parameters["thresh_trap_clogged"]
        ).mean()
        return frac_clogged_traps


def groupby_traps(traps, labels, edgemasks, ntraps):
    # Group data by traps to pass onto extractor without re-reading hdf5
    iterators = [
        groupby(zip(traps, dset), lambda x: x[0]) for dset in (labels, edgemasks)
    ]
    label_d = {key: [x[1] for x in group] for key, group in iterators[0]}
    mask_d = {
        key: np.dstack([ndimage.morphology.binary_fill_holes(x[1]) for x in group])
        for key, group in iterators[1]
    }
    labels = {i: label_d.get(i, []) for i in range(ntraps)}
    masks = {i: mask_d.get(i, []) for i in range(ntraps)}

    return labels, masks



import numpy as np
from inspect import getmembers, isfunction, getargspec
from extraction.core.functions import cell, trap
from extraction.core.functions.custom import localisation
from extraction.core.functions.math import div0
from extraction.core.functions.distributors import trap_apply


def load_cellfuns_core():
    # Generate str -> trap_function dict from functions in core.cell
    return {f[0]: f[1] for f in getmembers(cell) if isfunction(f[1])}


def load_custom_args():
    """
    Load custom functions. If they have extra arguments also load these
    """
    funs = {f[0]: f[1] for f in getmembers(localisation) if isfunction(f[1])}
    args = {
        k: getargspec(v).args[2:]
        for k, v in funs.items()
        if set(["cell_mask", "trap_image"]).intersection(getargspec(v).args)
    }

    return ({k: funs[k] for k in args.keys()}, {k: v for k, v in args.items() if v})


def load_cellfuns():
    # Generate str -> trap_function dict from core.cell and core.trap functions
    cell_funs = load_cellfuns_core()
    CELLFUNS = {}
    for k, f in cell_funs.items():
        if isfunction(f):

            def tmp(f):
                return lambda m, img: trap_apply(f, m, img)

            CELLFUNS[k] = tmp(f)
    return CELLFUNS


def load_trapfuns():
    TRAPFUNS = {f[0]: f[1] for f in getmembers(trap) if isfunction(f[1])}
    return TRAPFUNS


def load_funs():
    CELLFUNS = load_cellfuns()
    TRAPFUNS = load_trapfuns()

    return CELLFUNS, TRAPFUNS, {**TRAPFUNS, **CELLFUNS}


def load_redfuns():  # TODO make defining reduction functions more flexible
    RED_FUNS = {
        "np_max": np.maximum,
        "np_mean": np.mean,
        "np_median": np.median,
        "None": None,
    }
    return RED_FUNS


def load_mergefuns():
    MERGE_FUNS = {"div0": div0, "np_add": np.add}
    return MERGE_FUNS



"""This is an example module to show the structure."""
from typing import Union

import numpy as np
from PIL import Image


class localImageViewer:
    """
    This class is used to quickly access position images without tiling
    from image.h5 objects.
    """

    def __init__(self, h5file):
        """This class takes one parameter and is used to add one to that
        parameter.

        :param parameter: The parameter for this class
        """
        self._hdf = h5py.File(h5file)
        self.positions = list(self._hdf.keys())
        self.current_position = self.positions[0]
        self.parameter = parameter

    def plot_position(channel=0, tp=0, z=0, stretch=True):
        pixvals = self._hdf[self.current_position][channel, tp, ..., z]
        if stretch:
            minval = np.percentile(pixvals, 0.5)
            maxval = np.percentile(pixvals, 99.5)
            pixvals = np.clip(pixvals, minval, maxval)
            pixvals = ((pixvals - minval) / (maxval - minval)) * 255

        Image.fromarray(pixvals.astype(np.uint8))


from aliby.tile.tiler import Tiler, TilerParameters, TrapLocations
from agora.io.writer import load_attributes


import json

with open("/home/alan/Documents/dev/skeletons/server_info.json", "r") as f:
    # json.dump(
    #     {
    #         "host": "islay.bio.ed.ac.uk",
    #         "username": "upload",
    #         "password": "",
    #     },
    #     f,
    # )
    server_info = json.load(f)


import h5py
from aliby.io.omero import Image

from agora.io.cells import Cells


class remoteImageViewer:
    def __init__(self, fpath):
        with h5py.File(fpath, "r") as f:
            self.image_id = f.attrs.get("image_id", None) or 105146
        # trap_locs = TrapLocations.from_source(fpath)
        with Image(self.image_id, **server_info) as image:
            self.tiler = Tiler.from_hdf5(image, fpath)
        self.cells = Cells.from_source(fpath)
        # if parameters is None:
        #     parameters = TilerParameters.default()

        # with h5py.File(hdf, "r") as f:
        #     # image_id = f.attrs["omero_id"]
        #     image_id = 16543

    def get_position(self):
        pass

    def get_position_timelapse(self):
        pass

    @property
    def full(self):
        if not hasattr(self, "_full"):
            self._full = {}
        return self._full

    def get_tc(self, tp):
        with Image(self.image_id, **server_info) as image:
            self.tiler.image = image.data
            return self.tiler.get_tc(tp, riv.tiler.ref_channel)

    def get_trap_timepoints(self, trap_id, tps):
        with Image(self.image_id, **server_info) as image:
            self.tiler.image = image.data
            if set(tps).difference(self.full.keys()):
                tps = set(tps).difference(self.full.keys())
                for tp in tps:
                    self.full[tp] = self.tiler.get_traps_timepoint(
                        tp, channels=[self.tiler.ref_channel], z=[0]
                    )[:, 0, 0, ..., 0]
            requested_trap = {tp: self.full[tp] for tp in tps}

            return requested_trap

    def get_labeled_trap(self, trap_id, tps):
        imgs = self.get_trap_timepoints(trap_id, tps)
        imgs_list = [x[trap_id] for x in imgs.values()]
        outlines = [
            riv.cells.at_time(tp, kind="edgemask").get(trap_id, []) for tp in tps
        ]
        lbls = [riv.cells.labels_at_time(tp).get(trap_id, []) for tp in tps]
        lbld_outlines = [
            np.dstack([mask * lbl for mask, lbl in zip(maskset, lblset)]).max(axis=2)
            if len(lblset)
            else np.zeros_like(imgs_list[0]).astype(bool)
            for maskset, lblset in zip(outlines, lbls)
        ]
        outline_concat = np.concatenate(lbld_outlines, axis=1)
        img_concat = np.concatenate(imgs_list, axis=1)
        return outline_concat, img_concat


import matplotlib.pyplot as plt

# fpath = "/home/alan/Documents/dev/skeletons/data/2019_07_16_aggregates_CTP_switch_2_0glu_0_0glu_URA7young_URA8young_URA8old_01/2019_07_16_aggregates_CTP_switch_2_0glu_0_0glu_URA7young_URA8young_URA8old_01/URA8_old007.h5"
fpath = "/home/alan/Documents/dev/skeletons/data/2021_11_01_01_Raf_00/2021_11_01_01_Raf_00/d1134002.h5"
riv = remoteImageViewer(fpath)
# pos = riv.get_tc(0)
out, img = riv.get_labeled_trap(9, list(range(0, 30)))
out_bak = out
out = dilation(out).astype(float)
out[out == 0] = np.nan
plt.imshow(
    np.concatenate(np.array_split(img, 6, axis=1)),
    interpolation=None,
    cmap="Greys_r",
)
plt.imshow(
    np.concatenate(np.array_split(out, 6, axis=1)),
    cmap="Set1",
    interpolation=None,
)
plt.show()

concat = lambda a: np.concatenate([x for x in a])
add = lambda a: np.sum(a, axis=0)
# plt.imshow(add(roll(tmp[0], 10), np.roll(roll(tmp[1], 11), 6, axis=0)))
# plt.show()



from copy import copy


def reassign_mo_bud(mo_bud, trans):
    """
    FIXME:
    Update mother_bud dictionary using another dict with tracks joined

    input
    :param mo_bud: dict with mother's ids as keys and daughters' as values
    :param trans: dict of joint tracks where moved track -> static track

    output
    mo_bud with updated cell ids
    """

    val2lst = lambda x: [j for i in x.values() for j in i]

    bud_inter = set(val2lst(mo_bud)).intersection(trans.keys())

    # translate mothers and carry  untranslated daughters
    new_mb = copy(mo_bud)
    for origin, target in trans:
        if target in new_mb:
            new_mb[target] = new_mb.get(target, set()).union(new_mb[origin])
        del new_mb[origin]

    # translate daughters
    for origin, target in trans:
        if origin in bud_inter:  # Limit scope of searcg
            for mo, da in new_mb.items():
                if origin in da:
                    da.remove(origin)
                    new_mb[mo] = da.union(target)

    return new_mb



## Trap-wise calculations

import numpy as np


def imBackground(cell_masks, trap_image):
    '''
    :param cell_masks: (numpy 3d array) cells' segmentation mask
    :param trap_image: the image for the trap in which the cell is (all
    channels)
    '''
    if not len(cell_masks):
        cell_masks = np.zeros_like(trap_image)

    background = ~cell_masks.sum(axis=2).astype(bool)
    return (np.median(trap_image[np.where(background)]))



import numpy as np


def trap_apply(cell_fun, cell_masks, trap_image, **kwargs):
    """
    Apply a cell_function to a mask, trap_image pair

    :param cell_fun: function to apply to a cell (from extraction/cell.py)
    :param cell_masks: (numpy 3d array) cells' segmentation mask
    :param trap_image: the image for the trap in which the cell is (all
    channels)
    :**kwargs: parameters to pass if needed for custom functions
    """

    cells_iter = (*range(cell_masks.shape[2]),)
    return [cell_fun(cell_masks[..., i], trap_image, **kwargs) for i in cells_iter]


def reduce_z(trap_image, fun):
    # Optimise the reduction function if possible
    if isinstance(fun, np.ufunc):
        return fun.reduce(trap_image, axis=2)
    else:
        return np.apply_along_axis(fun, 2, trap_image)



from tqdm import tqdm

from omero.gateway import BlitzGateway

# Helper funs
def connect_omero():
    conn = BlitzGateway(*get_creds(), host='islay.bio.ed.ac.uk', port=4064)
    conn.connect()
    return conn

def get_creds():
   return('upload',
          '***REMOVED***', #OMERO Password
   )

def download_file(f):
    """
    Download file in chunks using FileWrapper object
    """
    desc = 'Downloading ' + f.getFileName() + \
        ' (' + str(round(f.getFileSize()/1000**2, 2)) + 'Mb)'

    down_file = bytearray()
    for c in tqdm(f.getFileInChunks(), desc=desc):
        down_file += c

    return down_file



# Example of argo experiment explorer
from aliby.utils.argo import Argo
from extraction.core.extractor import Extractor
from extraction.core.parameters import Parameters
from extraction.core.functions.defaults import get_params

argo = Argo()
argo.load()
# argo.channels("GFP")
argo.tags(["Alan"])
argo.complete()
# argo.cExperiment()
# argo.tiler_cells()

# params = Parameters(**get_params("batman_ph_dual_fast"))


# def try_extract(d):
#     try:
#         params = Parameters(**get_params("batman_ph_dual_fast"))
#         ext = Extractor(params, source=d.getId())
#         ext.load_tiler_cells()
#         ext.process_experiment()
#         print(d.getId(), d.getName(), "Experiment processed")
#         return True
#     except:
#         print(d.getId(), d.getName(), "Experiment not processed")

#         return False


# from multiprocessing.dummy import Pool as ThreadPool

# pool = ThreadPool(4)
# results = pool.map(try_extract, argo.dsets)
# import pickle

# with open("results.pkl", "wb") as f:
#     pickle.dump(results, f)



"""
Functions to process, filter and merge tracks.
"""

# from collections import Counter

from copy import copy
from typing import Union, List

import numpy as np
import pandas as pd

from scipy.signal import savgol_filter

# from scipy.optimize import linear_sum_assignment
# from scipy.optimize import curve_fit

from matplotlib import pyplot as plt


def load_test_dset():
    # Load development dataset to test functions
    return pd.DataFrame(
        {
            ("a", 1, 1): [2, 5, np.nan, 6, 8] + [np.nan] * 5,
            ("a", 1, 2): list(range(2, 12)),
            ("a", 1, 3): [np.nan] * 8 + [6, 7],
            ("a", 1, 4): [np.nan] * 5 + [9, 12, 10, 14, 18],
        },
        index=range(1, 11),
    ).T


def get_ntps(track: pd.Series) -> int:
    # Get number of timepoints
    indices = np.where(track.notna())
    return np.max(indices) - np.min(indices)


def get_tracks_ntps(tracks: pd.DataFrame) -> pd.Series:
    return tracks.apply(get_ntps, axis=1)


def get_avg_gr(track: pd.Series) -> int:
    """
    Get average growth rate for a track.

    :param tracks: Series with volume and timepoints as indices
    """
    ntps = get_ntps(track)
    vals = track.dropna().values
    gr = (vals[-1] - vals[0]) / ntps
    return gr


def get_avg_grs(tracks: pd.DataFrame) -> pd.DataFrame:
    """
    Get average growth rate for a group of tracks

    :param tracks: (m x n) dataframe where rows are cell tracks and
        columns are timepoints
    """
    return tracks.apply(get_avg_gr, axis=1)


def clean_tracks(tracks, min_len: int = 6, min_gr: float = 0.5) -> pd.DataFrame:
    """
    Clean small non-growing tracks and return the reduced dataframe

    :param tracks: (m x n) dataframe where rows are cell tracks and
        columns are timepoints
    :param min_len: int number of timepoints cells must have not to be removed
    :param min_gr: float Minimum mean growth rate to assume an outline is growing
    """
    ntps = get_tracks_ntps(tracks)
    grs = get_avg_grs(tracks)

    growing_long_tracks = tracks.loc[(ntps >= min_len) & (grs > min_gr)]

    return growing_long_tracks


def merge_tracks(tracks, drop=False, **kwargs) -> pd.DataFrame:
    """
    Join tracks that are contiguous and within a volume threshold of each other

    :param tracks: (m x n) dataframe where rows are cell tracks and
        columns are timepoints
    :param kwargs: args passed to get_joinable

    returns

    :joint_tracks: (m x n) Dataframe where rows are cell tracks and
        columns are timepoints. Merged tracks are still present but filled
        with np.nans.
    """

    # calculate tracks that can be merged until no more traps can be merged
    joinable_pairs = get_joinable(tracks, **kwargs)
    if joinable_pairs:
        tracks = join_tracks(tracks, joinable_pairs, drop=drop)
    joint_ids = get_joint_ids(joinable_pairs)

    return (tracks, joint_ids)


def get_joint_ids(merging_seqs) -> dict:
    """
    Convert a series of merges into a dictionary where
    the key is the cell_id of destination and the value a list
    of the other track ids that were merged into the key

    :param merging_seqs: list of tuples of indices indicating the
    sequence of merging events. It is important for this to be in sequential order

    How it works:

    The order of merging matters for naming, always the leftmost track will keep the id

    For example, having tracks (a, b, c, d) and the iterations of merge events:

    0 a b c d
    1 a b cd
    2 ab cd
    3 abcd

    We shold get:

    output {a:a, b:a, c:a, d:a}

    """
    targets, origins = list(zip(*merging_seqs))
    static_tracks = set(targets).difference(origins)

    joint = {track_id: track_id for track_id in static_tracks}
    for target, origin in merging_seqs:
        joint[origin] = target

    moved_target = [
        k for k, v in joint.items() if joint[v] != v and v in joint.values()
    ]

    for orig in moved_target:
        joint[orig] = rec_bottom(joint, orig)

    return {
        k: v for k, v in joint.items() if k != v
    }  # remove ids that point to themselves


def rec_bottom(d, k):
    if d[k] == k:
        return k
    else:
        return rec_bottom(d, d[k])


def join_tracks(tracks, joinable_pairs, drop=False) -> pd.DataFrame:
    """
    Join pairs of tracks from later tps towards the start.

    :param tracks: (m x n) dataframe where rows are cell tracks and
        columns are timepoints

    returns (copy)

    :param joint_tracks: (m x n) Dataframe where rows are cell tracks and
        columns are timepoints. Merged tracks are still present but filled
        with np.nans.
    :param drop: bool indicating whether or not to drop moved rows

    """

    tmp = copy(tracks)
    for target, source in joinable_pairs:
        tmp.loc[target] = join_track_pairs(tmp.loc[target], tmp.loc[source])

        if drop:
            tmp = tmp.drop(source)

    return tmp


def join_track_pairs(track1, track2):
    tmp = copy(track1)
    tmp.loc[track2.dropna().index] = track2.dropna().values

    return tmp


def get_joinable(tracks, smooth=False, tol=0.1, window=5, degree=3) -> dict:
    """
    Get the pair of track (without repeats) that have a smaller error than the
    tolerance. If there is a track that can be assigned to two or more other
    ones, it chooses the one with a lowest error.

    :param tracks: (m x n) dataframe where rows are cell tracks and
        columns are timepoints
    :param tol: float or int threshold of average (prediction error/std) necessary
        to consider two tracks the same. If float is fraction of first track,
        if int it is absolute units.
    :param window: int value of window used for savgol_filter
    :param degree: int value of polynomial degree passed to savgol_filter

    """

    tracks.index.names = [
        "pos",
        "trap",
        "cell",
    ]  # TODO remove this once it is integrated in the tracker
    # contig=tracks.groupby(['pos','trap']).apply(tracks2contig)
    clean = clean_tracks(tracks, min_len=window + 1, min_gr=0.9)  # get useful tracks
    contig = clean.groupby(["pos", "trap"]).apply(get_contiguous_pairs)
    contig = contig.loc[contig.apply(len) > 0]
    # candict = {k:v for d in contig.values for k,v in d.items()}

    # smooth all relevant tracks

    linear = set([k for v in contig.values for i in v for j in i for k in j])
    if smooth:  # Apply savgol filter TODO fix nans affecting edge placing
        savgol_on_srs = lambda x: non_uniform_savgol(x.index, x.values, window, degree)
        smoothed_tracks = clean.loc[linear].apply(savgol_on_srs, 1)
    else:
        smoothed_tracks = clean.loc[linear].apply(lambda x: np.array(x.values), axis=1)

    # fetch edges from ids TODO (IF necessary, here we can compare growth rates)
    idx_to_edge = lambda preposts: [
        (
            [get_val(smoothed_tracks.loc[pre], -1) for pre in pres],
            [get_val(smoothed_tracks.loc[post], 0) for post in posts],
        )
        for pres, posts in preposts
    ]
    edges = contig.apply(idx_to_edge)

    closest_pairs = edges.apply(get_vec_closest_pairs, tol=tol)

    # match local with global ids
    joinable_ids = [
        localid_to_idx(closest_pairs.loc[i], contig.loc[i]) for i in closest_pairs.index
    ]

    return [pair for pairset in joinable_ids for pair in pairset]


get_val = lambda x, n: x[~np.isnan(x)][n] if len(x[~np.isnan(x)]) else np.nan


def localid_to_idx(local_ids, contig_trap):
    """Fetch then original ids from a nested list with joinable local_ids

    input
    :param local_ids: list of list of pairs with cell ids to be joint
    :param local_ids: list of list of pairs with corresponding cell ids

    return
    list of pairs with (experiment-level) ids to be joint
    """
    lin_pairs = []
    for i, pairs in enumerate(local_ids):
        if len(pairs):
            for left, right in pairs:
                lin_pairs.append((contig_trap[i][0][left], contig_trap[i][1][right]))
    return lin_pairs


def get_vec_closest_pairs(lst: List, **kwags):
    return [get_closest_pairs(*l, **kwags) for l in lst]


def get_closest_pairs(pre: List[float], post: List[float], tol: Union[float, int] = 1):
    """Calculate a cost matrix the Hungarian algorithm to pick the best set of
    options

    input
    :param pre: list of floats with edges on left
    :param post: list of floats with edges on right
    :param tol: int or float if int metrics of tolerance, if float fraction

    returns
    :: list of indices corresponding to the best solutions for matrices

    """
    if len(pre) > len(post):
        dMetric = np.abs(np.subtract.outer(post, pre))
    else:
        dMetric = np.abs(np.subtract.outer(pre, post))
    # dMetric[np.isnan(dMetric)] = tol + 1 + np.nanmax(dMetric) # nans will be filtered
    # ids = linear_sum_assignment(dMetric)
    dMetric[np.isnan(dMetric)] = tol + 1 + np.nanmax(dMetric)  # nans will be filtered

    ids = solve_matrix(dMetric)
    if not len(ids[0]):
        return []

    norm = (
        np.array(pre)[ids[len(pre) > len(post)]] if tol < 1 else 1
    )  # relative or absolute tol
    result = dMetric[ids] / norm
    ids = ids if len(pre) < len(post) else ids[::-1]

    return [idx for idx, res in zip(zip(*ids), result) if res <= tol]


def solve_matrix(dMetric):
    """
    Solve cost matrix focusing on getting the smallest cost at each iteration.

    input
    :param dMetric: np.array cost matrix

    returns
    tuple of np.arrays indicating picks with lowest individual value
    """
    glob_is = []
    glob_js = []
    if (~np.isnan(dMetric)).any():
        tmp = copy(dMetric)
        std = sorted(tmp[~np.isnan(tmp)])
        while (~np.isnan(std)).any():
            v = std[0]
            i_s, j_s = np.where(tmp == v)
            i = i_s[0]
            j = j_s[0]
            tmp[i, :] += np.nan
            tmp[:, j] += np.nan
            glob_is.append(i)
            glob_js.append(j)

            std = sorted(tmp[~np.isnan(tmp)])

    return (np.array(glob_is), np.array(glob_js))


def plot_joinable(tracks, joinable_pairs, max=64):
    """
    Convenience plotting function for debugging and data vis
    """

    nx = 8
    ny = 8
    _, axes = plt.subplots(nx, ny)
    for i in range(nx):
        for j in range(ny):
            if i * ny + j < len(joinable_pairs):
                ax = axes[i, j]
                pre, post = joinable_pairs[i * ny + j]
                pre_srs = tracks.loc[pre].dropna()
                post_srs = tracks.loc[post].dropna()
                ax.plot(pre_srs.index, pre_srs.values, "b")
                # try:
                #     totrange = np.arange(pre_srs.index[0],post_srs.index[-1])
                #     ax.plot(totrange, interpolate(pre_srs, totrange), 'r-')
                # except:
                #     pass
                ax.plot(post_srs.index, post_srs.values, "g")

    plt.show()


def get_contiguous_pairs(tracks: pd.DataFrame) -> list:
    """
    Get all pair of contiguous track ids from a tracks dataframe.

    :param tracks: (m x n) dataframe where rows are cell tracks and
        columns are timepoints
    :param min_dgr: float minimum difference in growth rate from the interpolation
    """
    # indices = np.where(tracks.notna())

    mins, maxes = [
        tracks.notna().apply(np.where, axis=1).apply(fn) for fn in (np.min, np.max)
    ]

    mins_d = mins.groupby(mins).apply(lambda x: x.index.tolist())
    mins_d.index = mins_d.index - 1  # make indices equal
    maxes_d = maxes.groupby(maxes).apply(lambda x: x.index.tolist())

    common = sorted(set(mins_d.index).intersection(maxes_d.index), reverse=True)

    return [(maxes_d[t], mins_d[t]) for t in common]


# def fit_track(track: pd.Series, obj=None):
#     if obj is None:
#         obj = objective

#     x = track.dropna().index
#     y = track.dropna().values
#     popt, _ = curve_fit(obj, x, y)

#     return popt

# def interpolate(track, xs) -> list:
#     '''
#     Interpolate next timepoint from a track

#     :param track: pd.Series of volume growth over a time period
#     :param t: int timepoint to interpolate
#     '''
#     popt = fit_track(track)
#     # perr = np.sqrt(np.diag(pcov))
#     return objective(np.array(xs), *popt)


# def objective(x,a,b,c,d) -> float:
#     # return (a)/(1+b*np.exp(c*x))+d
#     return (((x+d)*a)/((x+d)+b))+c

# def cand_pairs_to_dict(candidates):
#     d={x:[] for x,_ in candidates}
#     for x,y in candidates:
#         d[x].append(y)
#     return d


def non_uniform_savgol(x, y, window, polynom):
    """
    Applies a Savitzky-Golay filter to y with non-uniform spacing
    as defined in x

    This is based on https://dsp.stackexchange.com/questions/1676/savitzky-golay-smoothing-filter-for-not-equally-spaced-data
    The borders are interpolated like scipy.signal.savgol_filter would do

    source: https://dsp.stackexchange.com/a/64313

    Parameters
    ----------
    x : array_like
        List of floats representing the x values of the data
    y : array_like
        List of floats representing the y values. Must have same length
        as x
    window : int (odd)
        Window length of datapoints. Must be odd and smaller than x
    polynom : int
        The order of polynom used. Must be smaller than the window size

    Returns
    -------
    np.array of float
        The smoothed y values
    """
    if len(x) != len(y):
        raise ValueError('"x" and "y" must be of the same size')

    if len(x) < window:
        raise ValueError("The data size must be larger than the window size")

    if type(window) is not int:
        raise TypeError('"window" must be an integer')

    if window % 2 == 0:
        raise ValueError('The "window" must be an odd integer')

    if type(polynom) is not int:
        raise TypeError('"polynom" must be an integer')

    if polynom >= window:
        raise ValueError('"polynom" must be less than "window"')

    half_window = window // 2
    polynom += 1

    # Initialize variables
    A = np.empty((window, polynom))  # Matrix
    tA = np.empty((polynom, window))  # Transposed matrix
    t = np.empty(window)  # Local x variables
    y_smoothed = np.full(len(y), np.nan)

    # Start smoothing
    for i in range(half_window, len(x) - half_window, 1):
        # Center a window of x values on x[i]
        for j in range(0, window, 1):
            t[j] = x[i + j - half_window] - x[i]

        # Create the initial matrix A and its transposed form tA
        for j in range(0, window, 1):
            r = 1.0
            for k in range(0, polynom, 1):
                A[j, k] = r
                tA[k, j] = r
                r *= t[j]

        # Multiply the two matrices
        tAA = np.matmul(tA, A)

        # Invert the product of the matrices
        tAA = np.linalg.inv(tAA)

        # Calculate the pseudoinverse of the design matrix
        coeffs = np.matmul(tAA, tA)

        # Calculate c0 which is also the y value for y[i]
        y_smoothed[i] = 0
        for j in range(0, window, 1):
            y_smoothed[i] += coeffs[0, j] * y[i + j - half_window]

        # If at the end or beginning, store all coefficients for the polynom
        if i == half_window:
            first_coeffs = np.zeros(polynom)
            for j in range(0, window, 1):
                for k in range(polynom):
                    first_coeffs[k] += coeffs[k, j] * y[j]
        elif i == len(x) - half_window - 1:
            last_coeffs = np.zeros(polynom)
            for j in range(0, window, 1):
                for k in range(polynom):
                    last_coeffs[k] += coeffs[k, j] * y[len(y) - window + j]

    # Interpolate the result at the left border
    for i in range(0, half_window, 1):
        y_smoothed[i] = 0
        x_i = 1
        for j in range(0, polynom, 1):
            y_smoothed[i] += first_coeffs[j] * x_i
            x_i *= x[i] - x[half_window]

    # Interpolate the result at the right border
    for i in range(len(x) - half_window, len(x), 1):
        y_smoothed[i] = 0
        x_i = 1
        for j in range(0, polynom, 1):
            y_smoothed[i] += last_coeffs[j] * x_i
            x_i *= x[i] - x[-half_window - 1]

    return y_smoothed



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['aliby',
 'aliby.io',
 'aliby.tile',
 'aliby.utils',
 'extraction',
 'extraction.core',
 'extraction.core.functions',
 'extraction.core.functions.custom',
 'extraction.examples']

package_data = \
{'': ['*'],
 'aliby': ['trap_templates/*'],
 'extraction.examples': ['pairs_data/*']}

install_requires = \
['aliby-agora>=0.2.12,<0.3.0',
 'aliby-baby>=0.1.3,<0.2.0',
 'dask>=2021.12.0,<2022.0.0',
 'h5py==2.10',
 'imageio==2.8.0',
 'jaliby-post>=0.1.9,<0.2.0',
 'numpy>=1.16.0,<2.0.0',
 'omero-py>=5.6.2',
 'opencv-python>=4.5.4,<5.0.0',
 'pandas>=1.3.5,<2.0.0',
 'pathos>=0.2.8,<0.3.0',
 'py-find-1st>=1.1.5,<2.0.0',
 'requests-toolbelt>=0.9.1,<0.10.0',
 'scikit-image>=0.19.1,<0.20.0',
 'scikit-learn==0.22.2.post1',
 'tqdm>=4.62.3,<5.0.0',
 'zeroc-ice==3.6.5']

setup_kwargs = {
    'name': 'aliby',
    'version': '0.1.15',
    'description': '',
    'long_description': None,
    'author': 'Alan Munoz',
    'author_email': 'alan.munoz@ed.ac.uk',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7.1,<3.11',
}


setup(**setup_kwargs)



# File with defaults for ease of use
from typing import Union
from pathlib import PosixPath, Path
import json


def exparams_from_meta(meta: Union[dict, PosixPath, str], extras=["ph"]):
    """
    Obtain parameters from metadata of hdf5 file
    """
    meta = meta if isinstance(meta, dict) else load_attributes(meta)
    base = {
        "tree": {"general": {"None": ["area", "volume", "eccentricity"]}},
        "multichannel_ops": {},
    }

    av_channels = {
        "Citrine",
        "GFP",
        "GFPFast",
        "mCherry",
        "pHluorin405",
        "Flavin",
        "Cy5",
        "mKO2",
    }

    default_reductions = {"np_max"}
    default_metrics = {
        "mean",
        "median",
        "imBackground",
        "max2p5pc",
        "max2p5pc_med",
        "max5px",
        "max5px_med",
    }
    default_rm = {r: default_metrics for r in default_reductions}

    av_flch = av_channels.intersection(meta["channels/channel"]).difference(
        {"Brightfield", "DIC", "BrightfieldGFP"}
    )

    for ch in av_flch:
        base["tree"][ch] = default_rm

    base["sub_bg"] = av_flch

    # Additional extraction
    if "ph" in extras and {"pHluorin405", "GFPFast"}.issubset(av_flch):

        sets = {
            b + a: (x, y)
            for a, x in zip(
                ["", "_bgsub"],
                (
                    ["GFPFast", "pHluorin405"],
                    ["GFPFast_bgsub", "pHluorin405_bgsub"],
                ),
            )
            for b, y in zip(["em_ratio", "gsum"], ["div0", "np_add"])
        }
        for i, v in sets.items():
            base["multichannel_ops"][i] = [
                *v,
                default_rm,
            ]

    return base


def load_attributes(file: str, group="/"):
    with h5py.File(file, "r") as f:
        meta = dict(f[group].attrs.items())
    return meta



import os
from pathlib import Path, PosixPath

from collections.abc import Iterable
import logging
from time import perf_counter

# from copy import copy
from typing import Union, List, Dict, Callable
from datetime import datetime

import numpy as np
import pandas as pd
from scipy.sparse import dok_matrix, vstack, issparse
from tqdm import tqdm


from extraction.core.functions.loaders import (
    load_funs,
    load_custom_args,
    load_redfuns,
    load_mergefuns,
)
from extraction.core.functions.defaults import exparams_from_meta
from extraction.core.functions.distributors import trap_apply, reduce_z
from extraction.core.functions.utils import depth

from agora.abc import ProcessABC, ParametersABC
from agora.io.writer import Writer, load_attributes
from agora.io.cells import Cells
from aliby.tile.tiler import Tiler

import matplotlib.pyplot as plt

CELL_FUNS, TRAPFUNS, FUNS = load_funs()
CUSTOM_FUNS, CUSTOM_ARGS = load_custom_args()
RED_FUNS = load_redfuns()
MERGE_FUNS = load_mergefuns()

# Assign datatype depending on the metric used
m2type = {"mean": np.float32, "median": np.ubyte, "imBackground": np.ubyte}


class ExtractorParameters(ParametersABC):
    """
    Base class to define parameters for extraction
    :tree: dict of depth n. If not of depth three tree will be filled with Nones
        str channel -> U(function,None) reduction -> str metric

    """

    def __init__(
        self,
        tree: Dict[Union[str, None], Dict[Union[Callable, None], List[str]]] = None,
        sub_bg: set = set(),
        multichannel_ops: Dict = {},
    ):

        self.tree = fill_tree(tree)

        self.sub_bg = sub_bg
        self.multichannel_ops = multichannel_ops

    @staticmethod
    def guess_from_meta(store_name: str, suffix="fast"):
        """
        Make a guess on the parameters using the hdf5 metadata

        Add anything as a suffix, default "fast"


        Parameters:
            store_name : str or Path indicating the results' storage.
            suffix : str to add at the end of the predicted parameter set
        """

        with h5py.open(store_name, "r") as f:
            microscope = f["/"].attrs.get("microscope")  # TODO Check this with Arin
        assert microscope, "No metadata found"

        return "_".join((microscope, suffix))

    @classmethod
    def default(cls):
        return cls({})

    @classmethod
    def from_meta(cls, meta):
        return cls(**exparams_from_meta(meta))


class Extractor(ProcessABC):
    """
    Base class to perform feature extraction.

    Parameters
    ----------
    parameters: core.extractor Parameters
        Parameters that include with channels, reduction and
            extraction functions to use.
    store: str
        Path to hdf5 storage file. Must contain cell outlines.
    tiler: pipeline-core.core.segmentation tiler
        Class that contains or fetches the image to be used for segmentation.
    """

    default_meta = {"pixel_size": 0.236, "z_size": 0.6, "spacing": 0.6}

    def __init__(
        self, parameters: ExtractorParameters, store: str = None, tiler: Tiler = None
    ):
        self.params = parameters
        if store:
            self.local = store
            self.load_meta()
        else:  # In case no h5 file is used, just use the parameters straight ahead
            self.meta = {"channel": parameters.to_dict()["tree"].keys()}
        if tiler:
            self.tiler = tiler
        self.load_funs()

    @classmethod
    def from_tiler(cls, parameters: ExtractorParameters, store: str, tiler: Tiler):
        return cls(parameters, store=store, tiler=tiler)

    @classmethod
    def from_img(cls, parameters: ExtractorParameters, store: str, img_meta: tuple):
        return cls(parameters, store=store, tiler=Tiler(*img_meta))

    # @classmethod
    # def from_store(cls, parameters: ExtractorParameters, store: str, img_meta: tuple):
    #     return cls(parameters, store=store, tiler=Tiler(*img_meta))

    @property
    def channels(self):
        if not hasattr(self, "_channels"):
            if type(self.params.tree) is dict:
                self._channels = tuple(self.params.tree.keys())

        return self._channels

    @property
    def current_position(self):
        return self.local.split("/")[-1][:-3]

    @property
    def group(self):  # Path within hdf5
        if not hasattr(self, "_out_path"):
            self._group = "/extraction/"
        return self._group

    @property
    def pos_file(self, store_name="store.h5"):
        if not hasattr(self, "_pos_file"):
            return self.local

    def load_custom_funs(self):
        """
        Load parameters of functions that require them from expt.
        These must be loaded within the Extractor instance because their parameters
        depend on their experiment's metadata.
        """
        funs = set(
            [
                fun
                for ch in self.params.tree.values()
                for red in ch.values()
                for fun in red
            ]
        )
        funs = funs.intersection(CUSTOM_FUNS.keys())
        ARG_VALS = {
            k: {k2: self.get_meta(k2) for k2 in v} for k, v in CUSTOM_ARGS.items()
        }
        # self._custom_funs = {trap_apply(CUSTOM_FUNS[fun],])
        self._custom_funs = {}
        for k, f in CUSTOM_FUNS.items():

            def tmp(f):
                return lambda m, img: trap_apply(f, m, img, **ARG_VALS.get(k, {}))

            self._custom_funs[k] = tmp(f)

    def load_funs(self):
        self.load_custom_funs()
        self._all_cell_funs = set(self._custom_funs.keys()).union(CELL_FUNS)
        self._all_funs = {**self._custom_funs, **FUNS}

    def load_meta(self):
        self.meta = load_attributes(self.local)

    def get_traps(
        self, tp: int, channels: list = None, z: list = None, **kwargs
    ) -> tuple:
        if channels is None:
            channel_ids = list(range(len(self.tiler.channels)))
        elif len(channels):
            channel_ids = [self.tiler.get_channel_index(ch) for ch in channels]
        else:
            channel_ids = None

        if z is None:
            z = list(range(self.tiler.shape[-1]))

        traps = (
            self.tiler.get_traps_timepoint(tp, channels=channel_ids, z=z, **kwargs)
            if channel_ids
            else None
        )

        return traps

    def extract_traps(
        self,
        traps: List[np.array],
        masks: List[np.array],
        metric: str,
        labels: List[int] = None,
    ) -> dict:
        """
        Apply a function for a whole position.

        :traps: List[np.array] list of images
        :masks: List[np.array] list of masks
        :metric:str metric to extract
        :labels: List[int] cell Labels to use as indices for output DataFrame
        :pos_info: bool Whether to add the position as index or not

        returns
        :d: Dictionary of dataframe
        """

        if labels is None:
            raise Warning("No labels given. Sorting cells using index.")

        cell_fun = True if metric in self._all_cell_funs else False

        idx = []
        results = []

        for trap_id, (mask_set, trap, lbl_set) in enumerate(
            zip(masks, traps, labels.values())
        ):
            if len(mask_set):  # ignore empty traps
                result = self._all_funs[metric](mask_set, trap)
                if cell_fun:
                    for lbl, val in zip(lbl_set, result):
                        results.append(val)
                        idx.append((trap_id, lbl))
                else:
                    results.append(result)
                    idx.append(trap_id)

        return (tuple(results), tuple(idx))

    def extract_funs(
        self, traps: List[np.array], masks: List[np.array], metrics: List[str], **kwargs
    ) -> dict:
        """
        Extract multiple metrics from a timepoint
        """
        d = {
            metric: self.extract_traps(
                traps=traps, masks=masks, metric=metric, **kwargs
            )
            for metric in metrics
        }

        return d

    def reduce_extract(
        self, traps: Union[np.array, None], masks: list, red_metrics: dict, **kwargs
    ) -> dict:
        """
        :param red_metrics: dict in which keys are reduction funcions and
        values are strings indicating the metric function
        :**kwargs: All other arguments, must include masks and traps.
        """

        reduced_traps = {}
        if traps is not None:
            for red_fun in red_metrics.keys():
                reduced_traps[red_fun] = [
                    self.reduce_dims(trap, method=RED_FUNS[red_fun]) for trap in traps
                ]

        d = {
            red_fun: self.extract_funs(
                metrics=metrics,
                traps=reduced_traps.get(red_fun, [None for _ in masks]),
                masks=masks,
                **kwargs,
            )
            for red_fun, metrics in red_metrics.items()
        }
        return d

    def reduce_dims(self, img: np.array, method=None) -> np.array:
        # assert len(img.shape) == 3, "Incorrect number of dimensions"

        if method is None:
            return img

        return reduce_z(img, method)

    def extract_tp(
        self,
        tp: int,
        tree: dict = None,
        tile_size: int = 117,
        masks=None,
        labels=None,
        **kwargs,
    ) -> dict:
        """
        :param tp: int timepoint from which to extract results
        :param tree: dict of dict {channel : {reduction_function : metrics}}
        :**kwargs: Must include masks and preferably labels.
        """

        if tree is None:
            tree = self.params.tree

        ch_tree = {ch: v for ch, v in tree.items() if ch != "general"}
        tree_chs = (*ch_tree,)

        cells = Cells.hdf(self.local)

        # labels
        if labels is None:
            raw_labels = cells.labels_at_time(tp)
            labels = {
                trap_id: raw_labels.get(trap_id, []) for trap_id in range(cells.ntraps)
            }

        # masks
        t = perf_counter()
        if masks is None:
            raw_masks = cells.at_time(tp, kind="mask")
            nmasks = len([y.shape for x in raw_masks.values() for y in x])
            # plt.imshow(np.dstack(raw_masks.get(1, [[]])).sum(axis=2))
            # plt.savefig(f"{tp}.png")
            # plt.close()
            logging.debug(f"Timing:nmasks:{nmasks}")
            logging.debug(f"Timing:MasksFetch:TP_{tp}:{perf_counter() - t}s")

            masks = {trap_id: [] for trap_id in range(cells.ntraps)}
            for trap_id, cells in raw_masks.items():
                if len(cells):
                    masks[trap_id] = np.dstack(np.array(cells)).astype(bool)

        masks = [np.array(v) for v in masks.values()]

        # traps
        traps = self.get_traps(tp, tile_size=tile_size, channels=tree_chs)

        self.img_bgsub = {}
        if self.params.sub_bg:
            bg = [
                ~np.sum(m, axis=2).astype(bool)
                if np.any(m)
                else np.zeros((tile_size, tile_size))
                for m in masks
            ]

        d = {}

        for ch, red_metrics in tree.items():
            img = None
            # ch != is necessary for threading
            if ch != "general" and traps is not None and len(traps):
                img = traps[:, tree_chs.index(ch), 0]

            d[ch] = self.reduce_extract(
                red_metrics=red_metrics, traps=img, masks=masks, labels=labels, **kwargs
            )

            if (
                ch in self.params.sub_bg and img is not None
            ):  # Calculate metrics with subtracted bg
                ch_bs = ch + "_bgsub"

                self.img_bgsub[ch_bs] = []
                for trap, maskset in zip(img, bg):

                    cells_fl = np.zeros_like(trap)

                    is_cell = np.where(maskset)
                    if len(is_cell[0]):  # skip calculation for empty traps
                        cells_fl = np.median(trap[is_cell], axis=0)

                    self.img_bgsub[ch_bs].append(trap - cells_fl)

                d[ch_bs] = self.reduce_extract(
                    red_metrics=ch_tree[ch],
                    traps=self.img_bgsub[ch_bs],
                    masks=masks,
                    labels=labels,
                    **kwargs,
                )

        # Additional operations between multiple channels (e.g. pH calculations)
        for name, (chs, merge_fun, red_metrics) in self.params.multichannel_ops.items():
            if len(
                set(chs).intersection(set(self.img_bgsub.keys()).union(tree_chs))
            ) == len(chs):
                imgs = [self.get_imgs(ch, traps, tree_chs) for ch in chs]
                merged = MERGE_FUNS[merge_fun](*imgs)
                d[name] = self.reduce_extract(
                    red_metrics=red_metrics,
                    traps=merged,
                    masks=masks,
                    labels=labels,
                    **kwargs,
                )

        del traps, masks
        return d

    def get_imgs(self, channel, traps, channels=None):
        """
        Returns the image from a correct source, either raw or bgsub


        :channel: str name of channel to get
        :img: ndarray (trap_id, channel, tp, tile_size, tile_size, n_zstacks) of standard channels
        :channels: List of channels
        """

        if channels is None:
            channels = (*self.params.tree,)

        if channel in channels:
            return traps[:, channels.index(channel), 0]
        elif channel in self.img_bgsub:
            return self.img_bgsub[channel]

    def run(self, tree=None, tps: List[int] = None, save=True, **kwargs) -> dict:

        if tree is None:
            tree = self.params.tree

        if tps is None:
            tps = list(range(self.meta["time_settings/ntimepoints"]))

        d = {}
        for tp in tps:
            new = flatten_nest(
                self.extract_tp(tp=tp, tree=tree, **kwargs),
                to="series",
                tp=tp,
            )

            for k in new.keys():
                n = new[k]
                d[k] = pd.concat((d.get(k, None), n), axis=1)

        for k in d.keys():
            indices = ["experiment", "position", "trap", "cell_label"]
            idx = (
                indices[-d[k].index.nlevels :]
                if d[k].index.nlevels > 1
                else [indices[-2]]
            )
            d[k].index.names = idx

            toreturn = d

        if save:
            self.save_to_hdf(toreturn)

        return toreturn

    def extract_pos(
        self, tree=None, tps: List[int] = None, save=True, **kwargs
    ) -> dict:

        if tree is None:
            tree = self.params.tree

        if tps is None:
            tps = list(range(self.meta["time_settings/ntimepoints"]))

        d = {}
        for tp in tps:
            new = flatten_nest(
                self.extract_tp(tp=tp, tree=tree, **kwargs),
                to="series",
                tp=tp,
            )

            for k in new.keys():
                n = new[k]
                d[k] = pd.concat((d.get(k, None), n), axis=1)

        for k in d.keys():
            indices = ["experiment", "position", "trap", "cell_label"]
            idx = (
                indices[-d[k].index.nlevels :]
                if d[k].index.nlevels > 1
                else [indices[-2]]
            )
            d[k].index.names = idx

            toreturn = d

        if save:
            self.save_to_hdf(toreturn)

        return toreturn

    def save_to_hdf(self, group_df, path=None):
        if path is None:
            path = self.local

        self.writer = Writer(path)
        for path, df in group_df.items():
            dset_path = "/extraction/" + path
            self.writer.write(dset_path, df)
        self.writer.id_cache.clear()

    def get_meta(self, flds):
        if not hasattr(flds, "__iter__"):
            flds = [flds]
        meta_short = {k.split("/")[-1]: v for k, v in self.meta.items()}
        return {f: meta_short.get(f, self.default_meta.get(f, None)) for f in flds}


### Helpers
def flatten_nest(nest: dict, to="series", tp: int = None) -> dict:
    """
    Convert a nested extraction dict into a dict of series
    :param nest: dict contained the nested results of extraction
    :param to: str = 'series' Determine output format, either list or  pd.Series
    :param tp: int timepoint used to name the series
    """

    d = {}
    for k0, v0 in nest.items():
        for k1, v1 in v0.items():
            for k2, v2 in v1.items():
                d["/".join((k0, k1, k2))] = (
                    pd.Series(*v2, name=tp) if to == "series" else v2
                )

    return d


def fill_tree(tree):
    if tree is None:
        return None
    tree_depth = depth(tree)
    if depth(tree) < 3:
        d = {None: {None: {None: []}}}
        for _ in range(2 - tree_depth):
            d = d[None]
        d[None] = tree
        tree = d
    return tree


class hollowExtractor(Extractor):
    """Extractor that only cares about receiving image and masks,
    used for testing.
    """

    def __init__(self, parameters):
        self.params = parameters
